// 2024 GDR Outro
//
// Coded fast and loose like I like my retro RPGs.
// Please never write production code like this.


#include "raylib.h"
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "fastprng.h"


#define SCREEN_W      1224
#define SCREEN_H       672
#define STAR_DENSITY  2000
#define PIXEL_DENSITY  365

#define TO_DEGREES(a) (a * (180.00 / PI))
#define TO_RADIANS(a) (a * (PI / 180.00))

#define C_BACKGROUND  (Color){  20,  28,  48, 255 }
#define C_WHITE_ZERO  (Color){ 255, 255, 255,   0 }
#define SKY_BLUE      (Color){  25, 113, 180, 255 }
#define CHILL_CYAN    (Color){  22, 230, 255, 255 }


// tracks pixel locations (exactly 365 virtual pixels)
// If you change the font this may FUBAR things
//
int totalizer = 0;
Vector2 pixels[PIXEL_DENSITY];

// Delta timing
//
float anims[6]     = { 0 };
float rotation     = 0.0;
float delta        = 0.0;
float delta_factor = 0.0;
float seconds      = 0.0;

// GDR Sine tracking
//
float g_angle  = 0.0;
float g_offset = 0.0;

// Scrolly Sine tracking
//
float scroll_x = 0.0;
float s_angle  = 0.0;
float s_offset = 0.0;

fastPRNG::fastXS64 fastR;                 // fastPRNG, because Raylib's is not so good


Vector4 stars[4][STAR_DENSITY];
Vector3 star_vars[4][STAR_DENSITY];


void  render_frame();
void  cleanup();

void  draw_letter(int letter, float x, float y, Color color);
void  eval_letter(int letter, float x, float y, float span);

int   box_dist(float x1, float y1, float x2, float y2, float radius);
void  draw_stars();
void  update_stars();

Vector2 advance(float angle, float velocity);
void push_rect(Rectangle *rect, float x, float y, float w, float h);
void tex_blit(Texture2D tex, int s_x, int s_y, int d_x, int d_y, int w, int h, float rotation, float scale, Color col);
float PR_rescale(float input, float in_min, float in_max, float out_min, float out_max);
int is_odd(int number);
int quick_rand(int range);
float quick_rand_f(float range);


///////////////////// = visible window
const char *scrolly[] = {
"                    WELCOME TO THE GDR 2024 OUTRO... WHERE WE CELEBRATE ALL THE STUFF WE WERE SUPPOSED TO FINISH THE YEAR BEFORE. WELCOME BACK TO OUR GOOD FRIENDS JONA AND ZHIL. CONGRATS TO EVERYONE FOR GETTING THROUGH THE 4X COMPOS THIS YEAR AND THE LIVE JAM.  GREETS GO TO ALL HOOPY FROODS....  0XDB    SONRISU    BEAN    POV    DELETER    MADGARDEN    CLIFFYA    BADMRBOX    HARD ROCK     WILL    GERED     KRONKSAN     EL DIABLO     SYN9     BENTACLE    ALEX     BEGUI     ZHIL     DARKDREAD     XEGNMA     MANDRAKE     CARRA     BRIGHAMMER     NIUNIO MARTINEZ          AND ANYONE ELSE I MAY HAVE FORGOTTEN....     2025    HERE. WE. GO....                                                                                                                                                              "
};


// Raylib-specific assets
//
Texture2D       t_star;
Image           i_font;
Music           muzak;
RenderTexture2D target;
Camera          camera;
Shader          crt;




int main(void)
{
 int i       = 0;
 int j       = 0;

 // Raylib init -- OpenGL context, background stuff, audio
 //
 SetConfigFlags(FLAG_VSYNC_HINT);

 InitWindow(SCREEN_W, SCREEN_H, "- GDR 2024 OUTRO -");
 InitAudioDevice();
 SetMasterVolume(0.50);


 // Load resources
 //
 t_star     = LoadTexture("gfx/t_star.png");
 crt        = LoadShader(0, "gfx/crt.fs");
 i_font     = LoadImage("gfx/t_font.png");
 muzak      = LoadMusicStream("sfx/music.mp3");
 target     = LoadRenderTexture(1224, 672);

 atexit(cleanup);


 // Launch music and set volume to 1/2
 //
 SetMusicVolume(muzak, 0.50);
 PlayMusicStream(muzak);



 // Initialize stars
 //
 for(i=0; i<4; i++) {
     for(j=0; j<STAR_DENSITY; j++) {
         stars[i][j].x = quick_rand_f((float)SCREEN_W);                   // X pos (unless overridden)
         stars[i][j].y = quick_rand_f((float)SCREEN_H);                   // Y pos (unless overridden)
         stars[i][j].z = 0.15 + (float)((float)quick_rand(75) / 100.0);   // Base scale
         stars[i][j].w = quick_rand(4);                                   // Parallax layer

         star_vars[i][j].x = quick_rand(360);                             // Roll Angle offset
         star_vars[i][j].y = 0.0;                                         // Excitation amount
         star_vars[i][j].z = 800.0 * sqrt(quick_rand_f(1));               // Radius for polar coords
                                   }
                    }




 // Main game loop
 //
 while (!WindowShouldClose()) {

     delta = GetFrameTime();
     seconds += delta;

     // Set up delta factor
     //
     delta_factor = ((float)delta / 0.0166666666666667);


     // Star rotation animation factor
     //
     anims[4] += 0.5 * delta_factor;
     if(anims[4] >= 360.00) { anims[4] -= 360.00; }


     // Scrolly x position tracker
     //
     scroll_x += (8.0 * delta_factor);
     if(scroll_x > 50000) { scroll_x -= 50000; }


     // Star parallax rotation factors
     //
     anims[0] += 0.03 * delta_factor;
     if(anims[0] >= 360.00) { anims[0] -= 360.00; }

     anims[1] += 0.04 * delta_factor;
     if(anims[1] >= 360.00) { anims[1] -= 360.00; }

     anims[2] += 0.05 * delta_factor;
     if(anims[2] >= 360.00) { anims[2] -= 360.00; }

     anims[3] += 0.06 * delta_factor;
     if(anims[3] >= 360.00) { anims[3] -= 360.00; }


     // Some sines
     //
     g_angle += (1.0 * delta_factor);
     if(g_angle > 360.00) { g_angle -= 360.00; }

     s_angle += (4.0 * delta_factor);
     if(s_angle > 360.00) { s_angle -= 360.00; }


     // Draw the rest of the Owl
     //
     update_stars();
     UpdateMusicStream(muzak);
     render_frame();
                              }

 return(0);
}






void render_frame()
{
 int i = 0;

 BeginTextureMode(target);

     // Clear FBO and give it a lick of paint
     //
     ClearBackground(C_BACKGROUND);
     DrawRectangle(0, 0, SCREEN_W, SCREEN_H, Fade(PURPLE, 0.25));

     // GDR Lettering
     //
     totalizer = 0;
     eval_letter(71, 50.0,  112.0, g_angle);
     eval_letter(68, 450.0, 112.0, g_angle + 45.0);
     eval_letter(82, 850.0, 112.0, g_angle + 90.0);

     draw_stars();

     // MAXIMUM LAZINESS FTW. Draw all the letters and let the GPU kick 'em out
     //
     for(i=0; i<660; i++) {
         draw_letter((int)scrolly[0][i], -scroll_x + (76.0 * i) + 10, 580.0, BLUE);
         draw_letter((int)scrolly[0][i], -scroll_x + (76.0 * i), 570.0, WHITE);
                          }

 EndTextureMode();



 // Apply CRT shader to FBO during final blit to video
 // Also, FBOs are Australian so you have to flip them over.
 // (See -target.texture.height, below)
 //
 BeginDrawing();

     BeginShaderMode(crt);
     DrawTextureRec(target.texture, (Rectangle){ 0, 0, (float)target.texture.width, (float)-target.texture.height }, (Vector2){ 0, 0 }, WHITE);
     EndShaderMode();

 EndDrawing();

 return;
}






// NOTE: This is automatically executed at program termination
//
void cleanup()
{

 UnloadTexture(t_star);
 UnloadImage(i_font);
 UnloadMusicStream(muzak);
 UnloadShader(crt);

 CloseWindow();

 return;
}




/// LETTER ROUTINES BELOW ---------------------------------------------------




// NOTE: Draws a letter, with each "pixel" exploded into a square.
//       The squares are drawn with empty space around each pixel.
//
// NOTE: 48-57 = 0-9
//       65-90 = A-Z
//
void draw_letter(int letter, float x, float y, Color color)
{
 Color c      = BLANK;
 int read_x   = 0;
 int read_y   = 0;
 int i        = 0;
 int j        = 0;
 int b_size   = 5;
 int b_space  = 5;
 float border = 100.0; // in pixels
 float alpha  = 1.0;

 div_t d;
 Vector2 pos    = { 0 };
 Vector2 adjust = { 0 };

 if(x < -50) { return; }
 if(x > GetScreenWidth()) { return; }


 // Individual glyphs are 15x13 with 1x1 whitespace
 //
 d = div(letter, 16);
 read_x = 1 + (15 * d.rem);
 read_y = 1 + (13 * d.quot);


 // Step through each "pixel" of the font glyph, and explode it into a square
 //
 for(i=0; i<13; i++) {
     pos.y = y + (b_space * i);

     for(j=0; j<15; j++) {
         c = GetImageColor(i_font, read_x + j, read_y + i);
         alpha = 1.0;

         pos.x = x + (b_space * j);

         adjust = advance(s_angle + pos.x, 20.0);
         pos.y = (y + (b_space * i)) + adjust.y;

         // Adjust alpha to fade out at edges of the screen
         //
         if(pos.x < border) { // Left edge detect
             alpha = PR_rescale(pos.x, 0.0, border, 0.0, 1.0);
                            }

         if(pos.x > (GetScreenWidth() - border)) { // Right edge detect
             alpha = PR_rescale(pos.x, (GetScreenWidth() - border), GetScreenWidth(), 1.0, 0.0);
                                                 }

         alpha *= PR_rescale(i,0,12, 1.0, 0.25);

         // Draw block if alpha shows visibility
         //
         if(c.a > 0.0) {
             DrawRectangle(pos.x, pos.y, b_size, b_size, Fade(color, alpha * (float)(c.a / 255.0)));
                       }
                         }
                     }

 return;
}




// NOTE: Same as above, but allows for multiple sine-wave tracking,
//       and these letters will not fade out at screen edges
//
void eval_letter(int letter, float x, float y, float span)
{
 Color c       = BLANK;
 int read_x    = 0;
 int read_y    = 0;
 int i         = 0;
 int j         = 0;
 float b_space = 24;
 float b_span  = 0.0;
 float b_start = 0.0;
 float alpha   = 0.25;
 div_t d;

 Vector2 pos   = { 0 };


 // Individual glyphs are 15x13 with 1x1 whitespace
 //
 d = div(letter, 16);
 read_x = 1 + (15 * d.rem);
 read_y = 1 + (13 * d.quot);

 // This was for an effect I didn't end up using
 //
 b_span  = 32;

 for(i=0; i<13; i++) {
     for(j=0; j<15; j++) {
         b_start = y + (sin(TO_RADIANS(span)) * 40.00);
         span += 0.5; // This is the magic

         // Original code
         c = GetImageColor(i_font, read_x + j, read_y + i);

         if(c.a > 0.0) {
             pos.x = x + (b_space * (float)j);
             pos.y = b_start + (b_span * (float)i);

             pixels[totalizer].x = pos.x;
             pixels[totalizer].y = pos.y;
             totalizer++;

             // This is wasteful, but gives the letters a faint
             // hint of solidity. Could be replaced with a circular
             // texture for less GPU load.
             //
             DrawRing(pos, 0, 10, 0, 360.00, 3.0, Fade(WHITE, alpha));
                       }
                         }
                     }

 return;
}



/// STAR ROUTINES BELOW -----------------------------------------------------



void draw_stars()
{
 int   i     = 0;
 int   j     = 0;
 float angle = 0.0;
 float alpha = 0.0;
 float scale = 0.0;

 // Larger stars
 //
 float scaling[4]   = { 1.20, 1.40, 1.60, 1.80 };
 float frequency[4] = { 1.00, 0.80, 0.50, 0.30 };
 float opacity[4]   = { 0.10, 0.15, 0.20, 0.25 };

 // Smaller stars
 //
 // float scaling[4]   = { 0.6, 0.65, 0.7, 0.75 };
 // float frequency[4] = { 1.0, 1.0, 1.0, 1.0 };
 // float opacity[4]   = { 0.10, 0.20, 0.35, 0.50 };

 Color colors[4]    = { WHITE, BLUE, CHILL_CYAN, SKY_BLUE };

 for(i=0; i<4; i++) {
     alpha = opacity[i];

     // First pass draws normally
     //
     for(j=0; j<(int)(STAR_DENSITY * frequency[i]); j++) {
         scale = stars[i][j].z;
         scale *= scaling[i];

         if(is_odd(j)) { angle = star_vars[i][j].x + anims[4]; }
         else          { angle = star_vars[i][j].x - anims[4]; }

         tex_blit(t_star, 0, 0, stars[i][j].x, stars[i][j].y, 64, 64, angle, scale, Fade(colors[(int)stars[i][j].w], alpha));
                                                         }

     // Second pass draws affected stars
     // Switching blend mode on/off repeatedly kills performance,
     // which is why this is split off
     //
     BeginBlendMode(BLEND_ADDITIVE);
     for(j=0; j<(int)(STAR_DENSITY * frequency[i]); j++) {
         if(star_vars[i][j].y > 0.0) {
             scale = stars[i][j].z;
             scale *= scaling[i];

             if(is_odd(j)) { angle = star_vars[i][j].x + anims[4]; }
             else          { angle = star_vars[i][j].x - anims[4]; }

             scale *= PR_rescale(star_vars[i][j].y, 0.0, 1.0, 1.0, 1.5);
             tex_blit(t_star, 0, 0, stars[i][j].x, stars[i][j].y, 64, 64, angle, scale, Fade(colors[(int)stars[i][j].w], star_vars[i][j].y * 0.50));

             scale *= PR_rescale(star_vars[i][j].y, 0.0, 1.0, 1.0, 1.7);
             tex_blit(t_star, 0, 0, stars[i][j].x, stars[i][j].y, 64, 64, angle, scale, Fade(colors[(int)stars[i][j].w], star_vars[i][j].y * 0.50));
                                     }
                                                         }
     EndBlendMode();
                    }

 return;
}




 // Update stars
 //
void update_stars()
{
 int i = 0;
 int j = 0;
 int k = 0;
 float star_decay   = 0.030;
 float star_excite  = 0.50;
 float frequency[4] = { 1.0, 1.0, 1.0, 1.0 };
 Vector2 dest       = { 0 };


 for(i=0; i<4; i++) {
     for(j=0; j<(int)(STAR_DENSITY * frequency[i]); j++) {

         // Decay stars that have been pulsed
         //
         if(star_vars[i][j].y > 0.0) {
             star_vars[i][j].y -= star_decay;
             if(star_vars[i][j].y < 0.0) { star_vars[i][j].y = 0.0; }
                                     }

         // Stars rotate around the center of the screen, using ersatz polar coords
         //
         if(is_odd(j)) { dest = advance(star_vars[i][j].x - anims[i], star_vars[i][j].z); }
         else          { dest = advance(star_vars[i][j].x + anims[i], star_vars[i][j].z); }

         stars[i][j].x = (SCREEN_W / 2.0) + dest.x;
         stars[i][j].y = (SCREEN_H / 2.0) + dest.y;

         // This is fucking lazy but IDGAF.
         // Do a quick box-collision scan of all GDR letter points.
         // A collision will excite a star.
         //
         for(k=0; k<PIXEL_DENSITY; k++) {
             if(box_dist(stars[i][j].x, stars[i][j].y, pixels[k].x, pixels[k].y, 14)) { star_vars[i][j].y = star_excite; }
                                        }
                                                         }
                    }

 return;
}



/// HELPER ROUTINES BELOW ---------------------------------------------------


float PR_rescale(float input, float in_min, float in_max, float out_min, float out_max)
{
 float result = 0.0;

 result = ((input - in_min) / (in_max - in_min)) * (out_max - out_min) + out_min;

 return(result);
}



int quick_rand(int range)
{
 // This has to be a double, or the lib shits itself
 //
 double f_range = range;
 float result   = fastR.xoroshiro128p_Range(0.0, f_range);

 return(result);
}



float quick_rand_f(float range)
{
 // This has to be a double, or the lib shits itself
 //
 double f_range = range;
 float result = fastR.xoroshiro128p_Range(0.0, f_range);

 return(result);
}



void tex_blit(Texture2D tex, int s_x, int s_y, int d_x, int d_y, int w, int h, float rotation, float scale, Color col)
{
 Rectangle source = { 0 };
 Rectangle dest   = { 0 };
 Vector2 origin   = { ((w / 2) * scale), ((h / 2) * scale) };

 push_rect(&source, s_x, s_y, w, h);
 push_rect(&dest, d_x, d_y, w * scale, h * scale);

 DrawTexturePro(tex, source, dest, origin, rotation, col);

 return;
}



// NOTE - Returns one if number is ODD; else zero.
//
int is_odd(int number)
{
 return (number & 1);
}



void push_rect(Rectangle *rect, float x, float y, float w, float h)
{

 rect->x      = x;
 rect->y      = y;
 rect->width  = w;
 rect->height = h;

 return;
}



Vector2 advance(float angle, float velocity)
{
 Vector2 result = { 0 };

 result.x = sin(TO_RADIANS(angle)) * velocity;
 result.y = -cos(TO_RADIANS(angle)) * velocity;

 return(result);
}



int box_dist(float x1, float y1, float x2, float y2, float radius)
{
 if((fabs((double)x1-(double)x2) < radius) && (fabs((double)y1-(double)y2) < radius)) { return(1); }
 return(0);
}
